#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
import os
import sys
import urllib
import os.path

sys.path.insert(0, os.path.abspath('../src/wa_kat'))
sys.path.insert(0, os.path.abspath('../src/'))
sys.path.insert(0, os.path.abspath('..'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinxcontrib.napoleon',
    'sphinx.ext.intersphinx'
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True

# Sorting of items
autodoc_member_order = "bysource"

# Document all methods in classes
autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'WA-KAT'
copyright = u'2015 Bystroushaak'

# The full version, including alpha/beta/rc tags.
try:
    # read data from CHANGES.rst
    sys.path.insert(0, os.path.abspath('../'))
    from docs import getVersion
    release = getVersion(open("../CHANGELOG.rst").read())
except:
    # this is here specially for readthedocs, which downloads only docs, not
    # other files
    fh = urllib.urlopen("https://pypi.python.org/pypi/" + project + "/")
    release = filter(lambda x: "<title>" in x, fh.read().splitlines())
    release = release[0].split(":")[0].split()[1]

# The short X.Y version.
version = ".".join(release.split(".")[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'default'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# Output file base name for HTML help builder.
htmlhelp_basename = 'WA-KAT'

# used to skip unwanted parts in some namedtuples
def maybe_skip_member(app, what, name, obj, skip, options):
    if what == "class" and name.startswith("_") or name in ["index", "count"]:
        return True
    return skip

def setup(app):
    app.connect('autodoc-skip-member', maybe_skip_member)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This file generates template with HTML / Brython application for the user.
"""
#
# Imports =====================================================================
import os
import json
import os.path

from bottle import get
from bottle import abort
from bottle import request
from bottle import template
from bottle import static_file

from shared import gzip_cache
from shared import read_template
from shared import in_template_path

from ..connectors import seeder


# Variables ===================================================================
INDEX_PATH = in_template_path("index_vertical.html")
STATIC_PATH = in_template_path("static")


# Functions & classes =========================================================
def read_index_template():
    """
    Open the template file and return it's content.

    Returns:
        str: Template content.
    """
    return read_template(INDEX_PATH)


def render_registered(url_id, remote_info):
    """
    Render template file for the registered user, which has some of the values
    prefilled.

    Args:
        url_id (str): Seeder URL id.
        remote_info (dict): Informations read from Seeder.

    Returns:
        str: Template filled with data.
    """
    return template(
        read_index_template(),
        registered=True,
        url=remote_info["url"],
        seeder_data=json.dumps(remote_info),
        url_id=url_id,
    )


def render_unregistered(error=None):
    """
    Render template file for the unregistered user.

    Args:
        error (str, default None): Optional error message.

    Returns:
        str: Template filled with data.
    """
    return template(
        read_index_template(),
        registered=False,
        error=error,
        seeder_data=None,
        url_id=None,
    )


# API =========================================================================
@get("/static/js/brython_dist.js")
def gzipped_brython():
    """
    Virtual file ``/static/js/brython_dist.js`` used as static GZIP cache to
    speed-up loading of `big` files.
    """
    path = os.path.join(STATIC_PATH, "js/brython_dist.js")
    return gzip_cache(path)


@get("/static/<fn:path>")
def static_data(fn):
    """
    Static file handler. This functions accesses all static files in ``static``
    directory.
    """
    file_path = os.path.normpath(fn)
    full_path = os.path.join(STATIC_PATH, file_path)

    if not os.path.exists(full_path):
        abort(404, "Soubor '%s' neexistuje!" % fn)

    return static_file(
        file_path,
        STATIC_PATH
    )


@get("/")
def render_form_template():
    """
    Rennder template for user.

    Decide whether the user is registered or not, pull remote info and so on.
    """
    error = ""
    remote_info = {}
    registered_user_id = request.query.get("url_id", False)

    # try to read remote info, the the url_id parameter was specified
    if registered_user_id:
        try:
            remote_info = seeder.get_remote_info(registered_user_id)
        except AssertionError:  #: TODO: requests error
            registered_user_id = False
            error = "Seeder neposlal očekávaná data.\n"

    if registered_user_id and remote_info:
        return render_registered(registered_user_id, remote_info)

    if not remote_info:
        error += "Seeder je nedostupný!\n"

    return render_unregistered(error)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from .log_view2 import LogView

from .errorbox import UrlBoxError
from .errorbox import ISSNBoxError

from .author_switcher import SWITCHER_EL
from .author_picker import AuthorPickerAdapter

from .progressbar import URLBar
from .progressbar import ISSNBar

from .output_picker import OutputPicker

from .keyword_handler import KeywordAdder
from .keyword_handler import UserKeywordHandler
from .keyword_handler import AlephKeywordHandler
from .keyword_handler import AanalysisKeywordHandler

from .input_controller import InputController

from .conspect_handler import ConspectDescriptor

from .dropdown_handler import DropdownHandler

from .placeholder_handler import PlaceholderHandler

from .shared import make_request
from .shared import func_on_enter



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================


# Functions & classes =========================================================
class PlaceholderHandler(object):
    """
    This class acts as handler for the background text, which is shown by
    bootstrap to blank elements.

    In case that there is multiple values in you typeahead tag, you may use
    this to show additional mesasge alerting the user to the possibility to
    choose from multiple values.
    """
    _dropdown_text = " Klikněte pro výběr."

    @staticmethod
    def set_placeholder_text(input_el, text):
        input_el.placeholder = text

    @staticmethod
    def get_placeholder_text(input_el):
        return input_el.placeholder

    @classmethod
    def set_placeholder_dropdown(cls, input_el):
        """
        Set the element to show the multiple choice text.
        """
        text = cls.get_placeholder_text(input_el)
        cls.set_placeholder_text(
            input_el=input_el,
            text=text + cls._dropdown_text
        )

    @classmethod
    def reset_placeholder_dropdown(cls, input_el):
        """
        Reset the element back to default.
        """
        text = cls.get_placeholder_text(input_el)
        cls.set_placeholder_text(
            input_el=input_el,
            text=text.replace(cls._dropdown_text, "")
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import json
import time
import random
import os.path
import tempfile
import threading
import subprocess

import pytest
import requests

from zeo_connector_defaults import tmp_context_name
from zeo_connector_defaults import generate_environment
from zeo_connector_defaults import cleanup_environment


# Variables ===================================================================
PORT = random.randint(20000, 60000)
URL = "http://127.0.0.1:%d" % PORT
API_URL = URL + "/api_v1/"
_SERVER_HANDLER = None


# Functions ===================================================================
def circuit_breaker_http_retry(max_retry=10):
    for i in range(max_retry):
        try:
            print "Connecting to server .. %d/%d" % (i + 1, max_retry)
            return requests.get(URL).raise_for_status()
        except Exception:
            time.sleep(1)

    raise IOError("Couldn't connect to thread with HTTP server. Aborting.")


def _create_alt_settings(path):
    alt_settings = {
        "ZEO_CLIENT_PATH": path,

        "WEB_ADDR": "127.0.0.1",
        "WEB_PORT": web_port(),
        "WEB_SERVER": "paste",
        "WEB_DEBUG": True,
        "WEB_RELOADER": True,
        "WEB_BE_QUIET": True,
    }

    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(json.dumps(alt_settings))
        return f.name


# Setup =======================================================================
@pytest.fixture(scope="session", autouse=True)
def zeo(request):
    generate_environment()
    request.addfinalizer(cleanup_environment)


@pytest.fixture(scope="session", autouse=True)
def client_conf_path(zeo):
    return tmp_context_name("zeo_client.conf")


@pytest.fixture(scope="session", autouse=True)
def web_port():
    return PORT


@pytest.fixture(scope="session", autouse=True)
def web_url():
    return URL


@pytest.fixture(scope="session", autouse=True)
def web_api_url():
    return API_URL


@pytest.fixture(scope="session", autouse=True)
def bottle_server(request, zeo, client_conf_path):
    alt_conf_path = _create_alt_settings(client_conf_path)

    # run the bottle REST server
    def run_bottle():
        # prepare path to the command
        command_path = os.path.join(
            os.path.dirname(__file__),
            "../bin/wa_kat_server.py"
        )
        assert os.path.exists(command_path)

        # replace settings with mocked file
        my_env = os.environ.copy()
        my_env["SETTINGS_PATH"] = alt_conf_path

        # run the server
        global _SERVER_HANDLER
        _SERVER_HANDLER = subprocess.Popen(command_path, env=my_env)

    serv = threading.Thread(target=run_bottle)
    serv.setDaemon(True)
    serv.start()

    # add finalizer which shutdowns the server and remove temporary file
    def shutdown_server():
        if _SERVER_HANDLER:
            _SERVER_HANDLER.terminate()
        os.unlink(alt_conf_path)

    request.addfinalizer(shutdown_server)

    # wait until the connection with server is created
    circuit_breaker_http_retry()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Worker definition for analyzers runner. Functions defined there are run as
standalone processes.
"""
#
# Imports =====================================================================
import time
import traceback

import transaction
from ZODB.POSException import ConflictError


# Functions & classes =========================================================
def _save_to_database(req, property_name, data):
    """
    Store `data` under `property_name` in the `req` database object.

    Args:
        req (obj): ZODB database object.
        property_name (str): Name of the property under which the `data` will
            be stored.
        data (obj): Any object.
    """
    with transaction.manager:
        setattr(req, property_name, data)

        req.processing_ended_ts = time.time()
        req._p_changed = True

    req.log("`%s` saved." % property_name)


def worker(url_key, property_name, function, function_arguments,
           conf_path=None):
    """
    This function usually runs as process on the background.

    It runs ``function(*function_arguments)`` and then stores them in ZODB
    in `url_key` under `property_name`.

    Warning:
        This function puts data into DB, isntead of returning them.

    Args:
        url_key (str): Key which will be used for database lookup.
        property_name (str): Name of the property used to store data.
        function (obj): Function used to load the data.
        function_arguments (list): List of parameters for function which will
            be called to retreive data.
        conf_path (str, default None): Optional parameter used by tests to
            redirect the database connections to test's environment.
    """
    # this may take some time, hence outside transaction manager
    error_msg = None
    try:
        data = function(*function_arguments)
    except Exception as e:
        data = []
        error_msg = "Error: " + traceback.format_exc().strip()
        error_msg += "\n" + str(e.message)

        with open("/home/bystrousak/wa_error.log", "a") as f:
            f.write(error_msg)

    # get the RequestInfo object from database
    from .request_database import RequestDatabase
    if conf_path:
        db = RequestDatabase(conf_path=conf_path)
    else:
        db = RequestDatabase()

    # save `data` into RequestInfo object property
    for i in xrange(5):
        # resolve the request_info object
        request_info = db.get_request(url_key)

        # handle logging of possible error message
        if error_msg:
            request_info.log(error_msg)
            error_msg = None

        func_name = str(function.__name__)
        request_info.log("Attempting to save output from `%s`." % func_name)

        # attempt to save the data into database
        try:
            return _save_to_database(
                req=request_info,
                property_name=property_name,
                data=data
            )
        except ConflictError:
            request_info.log("Attempt to save `%s` failed." % func_name)
            time.sleep(0.1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This part of the REST API joins the frontend dataset created by user with
convertors and connectors to various other parts of the WA-KAT backend, which
are used to fill missing data, and emit required output data.
"""
#
# Imports =====================================================================
import re
import json
import time
from os.path import join

from bottle import post
from bottle import SimpleTemplate
from bottle_rest import form_to_params

from marcxml_parser.tools.resorted import resorted

from ..settings import API_PATH

from ..convertors import to_dc
from ..convertors import mrc_to_marc
from ..convertors import item_to_mrc

from ..connectors.seeder import send_update

from shared import read_template

from keywords import keyword_to_info


# Functions & classes =========================================================
def compile_keywords(keywords):
    """
    Translate `keywords` to full keyword records as they are used in Aleph.

    Returns tuple with three lists, each of which is later used in different
    part of the MRC/MARC record.

    Args:
        keywords (list): List of keyword strings.

    Returns:
        tuple: (mdt_list, cz_keyword_list, en_keyword_list)
    """
    mdt = []
    cz_keywords = []
    en_keywords = []
    for keyword in keywords:
        keyword = keyword_to_info(keyword.encode("utf-8"))

        if not keyword:
            continue

        cz_keywords.append({
            "uid": keyword["uid"],
            "zahlavi": keyword["zahlavi"],
            "zdroj": "czenas",
        })

        if keyword.get("mdt"):
            mdt.append({
                "mdt": keyword["mdt"],
                "mrf": keyword["mrf"],
            })

        angl_ekvivalent = keyword.get("angl_ekvivalent")
        if angl_ekvivalent:
            en_keywords.append({
                "zahlavi": angl_ekvivalent,
                "zdroj": keyword.get("zdroj_angl_ekvivalentu") or "eczenas",
            })

    return mdt, cz_keywords, en_keywords


def render_mrc(data):
    """
    Render MRC template.

    Args:
        data (dict): Dictionary with data from the WA-KAT frontend.

    Returns:
        str: MRC template filled with `data`.
    """
    template_body = read_template("sablona_katalogizace_eperiodika.mrc")

    return SimpleTemplate(template_body).render(**data)


def url_to_fn(url):
    """
    Convert `url` to filename used to download the datasets.

    ``http://kitakitsune.org/xe`` -> ``kitakitsune.org_xe``.

    Args:
        url (str): URL of the resource.

    Returns:
        str: Normalized URL.
    """
    url = url.replace("http://", "").replace("https://", "")
    url = url.split("?")[0]

    return url.replace("%", "_").replace("/", "_")


def parse_date_range(date, alt_end_date=None):
    """
    Parse input `date` string in free-text format for four-digit long groups.

    Args:
        date (str): Input containing years.

    Returns:
        tuple: ``(from, to)`` as four-digit strings.
    """
    NOT_ENDED = "9999"
    all_years = re.findall(r"\d{4}", date)

    if alt_end_date:
        NOT_ENDED = alt_end_date

    if not all_years:
        return "****", NOT_ENDED

    elif len(all_years) == 1:
        return all_years[0], NOT_ENDED

    return all_years[0], all_years[1]


def serialize_author(author_data):
    """
    Author is passed as dictionary with parsed and raw data. This function adds
    *experimental support* for transport of all author fields based on RAW
    data, instead of parsed fields.

    If this proves as wrong, the code will be switched to use of parsed data
    again.

    Example of the input `author_data`::

       "author":{
          "name":"Grada Publishing",
          "code":"kn20080316009",
          "linked_forms":[
             "Grada Publishing a.s.",
             "Grada (nakladatelství)",
             "Nakladatelství Grada"
          ],
          "is_corporation":true,
          "record":{
             "7":[
                "kn20080316009"
             ],
             "i1":"2",
             "i2":" ",
             "a":[
                "Grada Publishing"
             ]
          },
          "alt_name":"Grada Publishing [organizace] (Grada Publishing a.s.,
                      Grada (nakladatelství), Nakladatelství Grada)"
        }

    This is transformed into::

        1102  L $$aGrada Publishing$$b $$7kn20080316009

    Fields used in the output are taken from the ``is_corporation``,
    ``record`` fields (``i1``, ``i2``, ``7`` and ``a``).

    Returns:
        str: String containing the line which will be included to the MRC.
    """
    record = author_data["record"]
    i1 = record["i1"]
    i2 = record["i2"]

    # make sure that all required fields will be created
    if "a" not in record:
        record["a"] = " "

    if author_data["is_corporation"]:
        code = "110"
        if "b" not in record:
            record["b"] = " "
    else:
        code = "100"
        if "d" not in record:
            record["d"] = " "
        if "4" not in record:
            record["4"] = " "

    out = "%s%s%s L " % (code, i1, i2)
    for key in resorted(key for key in record if len(key) == 1):
        for item in record[key]:
            out += "$$%s%s" % (key, item)

    return out


def get_008_lang_code(lang_code, default_lang="cze"):
    """
    008 lang code is always 3 characters long. Look at `language` info from the
    user's dataset and use it, if it is exactly 3 chars long.
    """
    if not lang_code:
        return default_lang

    if len(lang_code) != 3:
        return default_lang

    return lang_code


# REST API ====================================================================
@post(join(API_PATH, "to_output"))
@form_to_params
def to_output(data):
    """
    Convert WA-KAT frontend dataset to three output datasets - `MRC`, `MARC`
    and `Dublin core`.

    Conversion is implemented as filling ofthe MRC template, which is then
    converted to MARC record. Dublin core is converted standalone from the
    input dataset.
    """
    data = json.loads(data)

    # postprocessing
    if "keywords" in data:
        mdt, cz_keywords, en_keywords = compile_keywords(data["keywords"])
        del data["keywords"]

        data["mdt"] = mdt
        data["cz_keywords"] = cz_keywords
        data["en_keywords"] = en_keywords

    data["annotation"] = data["annotation"].replace("\n", " ")
    data["time"] = time  # for date generation

    # convert additional info values to MRC
    ai_key = "additional_info"
    if data.get(ai_key) is None:
        data[ai_key] = {}

    # parse regularity - see https://github.com/WebArchivCZ/WA-KAT/issues/66
    # for details
    fld_008 = data[ai_key].get("008")
    data["regularity"] = fld_008[18] if fld_008 and len(fld_008) > 18 else "-"

    # put lang code to 008
    data["lang_code_008"] = get_008_lang_code(data.get("language"))

    data[ai_key] = {
        key: "\n".join(item_to_mrc(key, val))
        for key, val in data[ai_key].iteritems()
        if val
    }

    alt_end_date = data[ai_key].get("alt_end_date", None)

    # handle date range in the 008
    from_year, to_year = parse_date_range(data["creation_date"], alt_end_date)
    data["from_year"] = from_year
    data["to_year"] = to_year

    # serialize author
    if data["author"]:
        data["serialized_author"] = serialize_author(data["author"])

    # send data to seeder
    if data.get("url_id"):
        send_update(data["url_id"], data)

    # convert to MRC format
    mrc = render_mrc(data).encode("utf-8")

    # create all output formats
    out = {
        "fn": url_to_fn(data["url"]),
        "mrc": mrc,
        "oai": mrc_to_marc(mrc),
        "dc": to_dc(data),
    }

    return out



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
Module with interface for picking authors from authority base in NK's Aleph.
"""
#
# Imports =====================================================================
import json
from os.path import join

from browser import html
from browser import alert
from browser import document

from .shared import make_request
from .shared import func_on_enter

from .log_view2 import LogView
from .progressbar import AuthorBar
from .author_switcher import SWITCHER_EL
from .overlay_controller import BlackOverlay

# virtual filesystem / modules provided by REST API
from virtual_fs import settings


# Functions & classes =========================================================
class AuthorPicker(object):
    """
    This component is responsible for generating the interface for picking
    authors from Aleph.
    """
    original_author_el = document["author"]

    div_el = document["author_picker"]  # whole <div>
    input_el = document["author_input"]  # <input> for author's name
    search_btn_el = document["search_author"]  # trigger the search of author
    select_el = document["author_select"]  # <select> element with authors
    storno_btn_el = document["storno_author"]  # used to storno the picking
    pick_btn_el = document["pick_author"]  # used to accept the selected author

    overlay = BlackOverlay()

    code_to_data = {}  # this is used to translate select back to data
    selected_code = None

    @classmethod
    def show(cls):
        """
        Show the author picker.
        """
        cls.div_el.style.display = "block"
        cls.hide_errors()

        cls.overlay.show()
        cls.overlay.el.bind("click", lambda x: cls.hide())

    @classmethod
    def hide(cls):
        """
        Hide the author picker.
        """
        cls.overlay.hide()
        cls.div_el.style.display = "none"

    @classmethod
    def hide_errors(cls):
        """
        Hide errors shown by validators.
        """
        cls.select_el.style.border = "0"
        cls.input_el.style.border = "0"

    @classmethod
    def _pick_selected_option(cls):
        """
        Select handler for authors.
        """
        for option in cls.select_el:
            # if the select is empty
            if not hasattr(option, "selected"):
                return None

            if option.selected:
                return option.value

        return None

    @classmethod
    def on_pick_button_pressed(cls, ev):
        """
        Callback called when the user press the button for accepting the picked
        author.

        This element calls validations, before accepting the choice.
        """
        cls.hide_errors()

        selected_code = cls._pick_selected_option()
        if not selected_code:
            cls.select_el.style.border = "2px solid red"
            return

        cls.selected_code = selected_code
        cls.code = str(selected_code)
        cls.original_author_el.value = cls.code_to_data[selected_code].get(
            "alt_name",
            str(cls.code_to_data[selected_code])
        )
        cls.hide()

    @classmethod
    def set_select(cls, authors):
        """
        Put data into ``<select>`` element.

        Args:
            authors (dict): Dictionary with author informations returned from
                aleph REST API. Format:
                ``{"name": .., "code": .., "linked_forms": ["..",]}``.
        """
        cls.select_el.html = ""

        if not authors:
            cls.select_el.disabled = True
            cls.select_el <= html.OPTION("Nic nenalezeno!")
            return

        cls.select_el.disabled = False
        for author_dict in authors:
            name = author_dict.get("name")
            code = author_dict.get("code")
            alt_name = author_dict.get("alt_name", name)

            if not (name and code):
                continue

            cls.code_to_data[code] = author_dict
            cls.select_el <= html.OPTION(alt_name, value=code)

    @classmethod
    def bind(cls):
        """
        Bind the callbacks to the buttons.
        """
        document["show_author_picker"].bind("click", lambda x: cls.show())

        cls.storno_btn_el.bind("click", lambda x: cls.hide())
        cls.pick_btn_el.bind("click", cls.on_pick_button_pressed)

    @classmethod
    def __get__(cls, obj, objtype):
        if not SWITCHER_EL.checked:
            return None

        return cls.code_to_data[cls.selected_code]

    @classmethod
    def __set__(cls, obj, value):
        # called by reset()
        if not value:
            cls.selected_code = None
            cls.original_author_el.value = ""
            return

        if type(value) in [list, tuple]:
            value = value[0]

        code = value["code"]
        cls.code_to_data[code] = value
        cls.selected_code = code

        cls.original_author_el.value = cls.code_to_data[code].get(
            "alt_name",
            str(cls.code_to_data[code])
        )

    @classmethod
    def validate(cls):
        """
        Validate required elements.
        """
        if not SWITCHER_EL.checked:
            return True

        if not cls.selected_code:
            cls.original_author_el.style.border = "2px solid red"
            return False

        cls.original_author_el.style.border = "0"
        return True


class AuthorPickerAdapter(AuthorPicker):
    """
    Adapter of the :class:`AuthorPicker` to aleph REST API.

    This adapter is mapped to the key / button press. It then launches querying
    of the REST API in periodic interval, updating the progress bar and so on.
    """
    @classmethod
    def on_complete(cls, req):
        """
        Callback called when the request was received.
        """
        # handle http errors
        if not (req.status == 200 or req.status == 0):
            alert("Couldn't connect to authority base.")
            LogView.add(
                "Error when calling Aleph authority base: %s" % req.text
            )
            AuthorBar.hide()
            return

        AuthorBar.hide()
        cls.set_select(json.loads(req.text))

    @classmethod
    def start(cls, ev):
        """
        Event handler which starts the request to REST API.
        """
        # somehow the first call doesn't stop the propagation
        ev.stopPropagation()
        ev.preventDefault()

        # make sure, that `author` was filled
        author = cls.input_el.value.strip()
        if not author:
            cls.input_el.style.border = "2px solid red"
            return

        cls.hide_errors()

        AuthorBar.show(50)
        make_request(
            url=join(settings.API_PATH, "aleph/authors_by_name"),
            data={"name": author},
            on_complete=cls.on_complete
        )

    @classmethod
    def bind(cls):
        """
        Bind the buttons to adapter's event handler.
        """
        super(cls, cls).bind()

        cls.search_btn_el.bind("click", cls.start)
        cls.input_el.bind("keypress", func_on_enter(cls.start))


AuthorPickerAdapter.bind()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This module is used as connector to Seeder - currator's application, which may
act as a source for some of the metadata.
"""
#
# Imports =====================================================================
import sys

import requests

from .. import settings
from ..data_model import Model
from ..analyzers.source_string import SourceString


# Functions & classes =========================================================
def _convert_to_wakat_format(seeder_struct):
    """
    Convert Seeder's structure to the internal structure used at frontend.

    Args:,
        seeder_struct (dict): Dictionary with Seeder data.

    Returns:
        obj: :class:`Model`.
    """
    def pick_active(seeder_struct, what):
        """
        From the list of dicts, choose only first of such, that contains
        ``"active": True`` item.

        If not found, just pick the first.

        Args:
            seeder_struct (dict): Dict with bunch of data.
            what (str): What key to use in `seeder_struct` to identify the
                list of dicts.

        Returns:
            dict: Active or first dict.
        """
        items = seeder_struct.get(what)

        if not items:
            return None

        if not (isinstance(items, list) or isinstance(items, tuple)):
            items = [items]

        active_items = [item for item in items if item.get("active")]

        if not active_items:
            return items[0]

        return active_items[0]

    if not seeder_struct:
        return None

    # pick active seed and active publisher
    active_seed = pick_active(seeder_struct, "seeds")
    publisher_contact = pick_active(
        seeder_struct.get("publisher", {}),
        "contacts"
    )

    # seed contains `url`, which is used as primary key - if no seed is found,
    # it is meaningless to continue
    if not active_seed:
        active_seed = pick_active(seeder_struct, "seed")  # alt naming
        if not active_seed:
            return None

    # create the model and fill it with data
    model = Model()
    model.url = active_seed["url"]
    model.issn = seeder_struct.get("issn")
    model.title_tags = seeder_struct.get("name")
    model.publisher_tags = seeder_struct.get("publisher", {}).get("name")
    model.annotation_tags = seeder_struct.get("comment")  # annotation?

    # conspect = None  # TODO: !

    if publisher_contact:
        model.place_tags = publisher_contact.get("address")

    # rules are stored in custom subdictionary
    model.rules = {}
    model.rules["frequency"] = str(seeder_struct.get("frequency"))

    # add source info
    for key in model.keys():
        val = getattr(model, key)
        if val and "tags" in key:
            setattr(model, key, [{"val": val, "source": "Seeder"}])

    return model.get_mapping()


def _send_request(url_id, data=None, json=None, req_type=None):
    """
    Send request to Seeder's API.

    Args:
        url_id (str): ID used as identification in Seeder.
        data (obj, default None): Optional parameter for data.
        json (obj, default None): Optional parameter for JSON body.
        req_type (fn, default None): Request method used to send/download the
            data. If none, `requests.get` is used.

    Returns:
        dict: Data from Seeder.
    """
    url = settings.SEEDER_INFO_URL % url_id

    if not req_type:
        req_type = requests.get

    resp = req_type(
        url,
        data=data,
        json=json,
        timeout=settings.SEEDER_TIMEOUT,
        headers={
            "User-Agent": settings.USER_AGENT,
            "Authorization": settings.SEEDER_TOKEN,
        }
    )
    resp.raise_for_status()
    data = resp.json()

    return data


def get_remote_info(url_id):
    """
    Download data and convert them to dict used in frontend.

    Args:
        url_id (str): ID used as identification in Seeder.

    Returns:
        dict: Dict with data for frontend or None in case of error.
    """
    try:
        data = _send_request(url_id)
    except Exception as e:
        sys.stderr.write("Seeder GET error: ")  # TODO: better!
        sys.stderr.write(str(e.message))
        return None

    return _convert_to_wakat_format(data)


def _convert_to_seeder_format(dataset):
    """
    WA KAT dataset has different structure from Seeder. This is convertor
    which converts WA-KAT -> Seeder data format.

    Args:
        dataset (dict): WA-KAT dataset sent from frontend.

    Returns:
        dict: Dict with converted data.
    """
    data = {}
    seed = {}

    def add_if_set(d, key, val):
        if val:
            d[key] = val

    add_if_set(data, "name", dataset.get("title"))
    add_if_set(data, "issn", dataset.get("issn"))
    add_if_set(data, "annotation", dataset.get("annotation"))

    rules = dataset.get("rules", {})
    if rules:
        add_if_set(data, "frequency", rules.get("frequency"))

        # set seed info
        add_if_set(seed, "budget", rules.get("budget"))
        add_if_set(seed, "calendars", rules.get("calendars"))
        add_if_set(seed, "global_reject", rules.get("global_reject"))
        add_if_set(seed, "gentle_fetch", rules.get("gentle_fetch"))
        add_if_set(seed, "javascript", rules.get("javascript"))
        add_if_set(seed, "local_traps", rules.get("local_traps"))
        add_if_set(seed, "youtube", rules.get("youtube"))

        add_if_set(seed, "url", dataset.get("url"))

    if seed:
        data["seed"] = seed

    return data


def send_update(url_id, dataset):
    """
    Send request to Seeder's API with data changed by user.

    Args:
        url_id (str): ID used as identification in Seeder.
        dataset (dict): WA-KAT dataset sent from frontend.
    """
    data = _convert_to_seeder_format(dataset)

    if not data:
        return

    try:
        _send_request(url_id, json=data, req_type=requests.patch)
    except Exception as e:
        sys.stderr.write("Seeder PATCH error: ")  # TODO: better!
        sys.stderr.write(str(e.message))
        return None



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import time

import transaction
from zeo_connector import transaction_manager
from zeo_connector.examples import DatabaseHandler

from ..settings import PROJECT_KEY
from ..settings import ZEO_CLIENT_PATH

from request_info import RequestInfo


# Variables ===================================================================
DAY = 60 * 60 * 24  #: Amount of seconds in one day.
YEAR = DAY * 356  #: Amount of seconds in one year.


# Functions & classes =========================================================
class RequestDatabase(DatabaseHandler):
    """
    Small database handler, which lets keeps track of currently running
    analysis. It also caches the requests, so it doesn't use that much
    resources.

    Attributes:
        request_key (str): Key which is used to access the ZODB.
        requests (OOTreeSet): Tree dictionary-like object under which the
            values are stored.
    """
    def __init__(self, conf_path=ZEO_CLIENT_PATH, project_key=PROJECT_KEY):
        """
        Constructor.

        Args:
            conf_path (str): Path to the ZEO client configuration file. Default
                :attr:`.ZEO_CLIENT_PATH`.
            project_key (str): Key which is used for whole DB. Default
                :attr:`.PROJECT_KEY`.
        """
        super(self.__class__, self).__init__(
            conf_path=conf_path,
            project_key=project_key
        )

        self.request_key = "requests"
        self.requests = self._get_key_or_create(self.request_key)

    def get_request(self, url, new=False):
        """
        For given `url` register new :class:`RequestInfo` object or return
        cached, if there is such.

        Args:
            url (str): Key for database under which the :class:`.RequestInfo`
                is stored.

        Returns:
            obj: Proper :class:`.RequestInfo` object.
        """
        with transaction.manager:
            old_req = self.requests.get(url, None)

        # return cached requests
        if old_req and not new:
            return old_req

        if not (url.startswith("http://") or url.startswith("https://")):
            raise ValueError("Invalid URL `%s`!" % url)

        # if not found, create new
        req = RequestInfo(url=url)

        with transaction.manager:
            if old_req:
                del self.requests[url]
                self.zeo.pack()

            self.requests[url] = req

        return req

    @transaction_manager
    def garbage_collection(self, time_limit=YEAR/2.0):
        """
        Collect and remove all :class:`.RequestInfo` objects older than
        `time_limit` (in seconds).

        Args:
            time_limit (float, default YEAR / 2): Collect objects older than
                this limit.
        """
        expired_ri_keys = [
            key
            for key, ri in self.requests.iteritems()
            if ri.creation_ts + time_limit <= time.time()
        ]

        for key in expired_ri_keys:
            del self.requests[key]

        self.zeo.pack()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.keyword_detector import get_keyword_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML>
<head>
    <title>HTML title</title>

    <meta name    = "keywords"
          content = "html, keywords">

    <meta name    = "DC.Keywords"
          content = "first second third">
</head>
<body>
Zde je nějaký obsah, ze kterého by se měly správně vyextrahovat klíčová slova,
jako třeba azgabash, nebo hardware. Naopak běžná slova by měla být ignorována.

Tyto slova je pak třeba porovnat ještě s nějakým setem podporovaných slov v
národní knihovně.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_keyword_tags():
    keywords = get_keyword_tags(TEST_TEMPLATE, map_to_nk_set=False)

    assert keywords[0] == "html"
    assert keywords[0].source == "HTML"

    assert keywords[1] == "keywords"
    assert keywords[1].source == "HTML"

    assert keywords[2] == "first"
    assert keywords[2].source == "DC"

    assert keywords[3] == "second"
    assert keywords[3].source == "DC"

    assert keywords[4] == "third"
    assert keywords[4].source == "DC"

    # TODO: free text keyword extraction
    # assert keywords[5] == "azgabash"
    # assert keywords[5].source == "text"

    # assert keywords[6] == "hardware"
    # assert keywords[6].source == "text"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Parse and guess information about language of the resource. Normalize the
language tags to ISO 639-2 format.
"""
#
# Imports =====================================================================
import langdetect
import dhtmlparser

from .shared import parse_meta
from .source_string import SourceString

from ..convertors.iso_codes import normalize


# Functions & classes =========================================================
def get_html_lang_tags(index_page):
    """
    Return `languages` stored in ``<meta>`` tags.

    ``<meta http-equiv="Content-language" content="cs">`` -> ``cs``

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    lang_tag = "content-language"
    lang_tags = dom.find(
        "meta",
        fn=lambda x: x.params.get("http-equiv", "").lower() == lang_tag
    )

    return [
        SourceString(tag.params["content"], "HTML")
        for tag in lang_tags
        if "content" in tag.params
    ]


def get_dc_lang_tags(index_page):
    """
    Return `languages` stored in dublin core ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "dc.language", "DC")


def detect_language(index_page):
    """
    Detect `languages` using `langdetect` library.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        obj: One :class:`.SourceString` object.
    """
    dom = dhtmlparser.parseString(index_page)

    clean_content = dhtmlparser.removeTags(dom)

    lang = None
    try:
        lang = langdetect.detect(clean_content)
    except UnicodeDecodeError:
        lang = langdetect.detect(clean_content.decode("utf-8"))

    return SourceString(
        lang,
        source="langdetect"
    )


def get_lang_tags(index_page):
    """
    Collect informations about language of the page from HTML and Dublin core
    tags and langdetect guesses.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    dom = dhtmlparser.parseString(index_page)

    titles = [
        get_html_lang_tags(dom),
        get_dc_lang_tags(dom),
        [detect_language(dom)],
    ]

    return list({
        SourceString(normalize(lang), source=lang.source)
        for lang in sum(titles, [])
    })



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.annotation_detector import get_annotation_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML>
<head>
    <meta name="description"
          content="Popis stránek.">.

    <meta name    = "DC.Description"
          content = "Description of the web.">
</head>
<body>
Somecontent.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_annotation_tags():
    descriptions = get_annotation_tags(TEST_TEMPLATE)

    assert descriptions[0] == "Popis stránek."
    assert descriptions[0].source == "Meta"

    assert descriptions[1] == "Description of the web."
    assert descriptions[1].source == "DC"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from browser import document

from .overlay_controller import BlackOverlay


# Functions & classes =========================================================
class LogView(object):
    """
    This class allows logging of states of the application AND generates a
    simple GUI for browsing this logs.
    """
    _value = []
    el = document["log_placeholder"]
    show_button = document["show_log"]

    overlay = BlackOverlay()

    @classmethod
    def _render(cls):
        cls.el.text = cls.get()

    @classmethod
    def add(cls, msg):
        """
        Add new `msg` to the log.
        """
        cls._value.append(msg)
        cls._render()

    @classmethod
    def get(cls):
        """
        Get all the log messages.
        """
        return "\n---\n".join(cls._value)

    @classmethod
    def show(cls, msg=None):
        """
        Show the log interface on the page.
        """
        if msg:
            cls.add(msg)

        cls.overlay.show()
        cls.overlay.el.bind("click", lambda x: cls.hide())

        cls.el.style.display = "block"
        cls.bind()

    @classmethod
    def hide(cls):
        """
        Hide the log interface.
        """
        cls.el.style.display = "none"
        cls.overlay.hide()
        cls.bind()

    @classmethod
    def bind(cls):
        """
        Bind the buttons to the callbacks.
        """
        cls.show_button.bind("click", lambda ev: cls.show())

LogView.bind()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from browser import document


# Functions & classes =========================================================
class DropdownHandler(object):
    """
    This class is used to "decorate" the inputs / selects with glyph pictograms
    used by bootstrap.
    """
    @staticmethod
    def _get_dropdown_glyph_el(input_id):
        input_el = document[input_id]
        parent = input_el.parent
        grand_parent = parent.parent

        for el in list(parent.children) + list(grand_parent.children):
            if el.class_name and "dropdown_hint" in el.class_name.lower():
                return el

        raise ValueError("Dropdown with id `%s` not found!" % input_id)

    @classmethod
    def show_dropdown_glyph(cls, input_id):
        el = cls._get_dropdown_glyph_el(input_id)
        el.style.display = "block"

        return el

    @classmethod
    def hide_dropdown_glyph(cls, input_id):
        try:
            el = cls._get_dropdown_glyph_el(input_id)
            el.style.display = "none"
        except (ValueError, KeyError):
            return

    @classmethod
    def set_dropdown_glyph(cls, input_id, glyph_name):
        el = cls.show_dropdown_glyph(input_id)
        filtered_tokens = [
            token
            for token in str(el.class_name).split()
            if "glyphicon" not in token
        ]
        tokens = filtered_tokens + ["glyphicon", glyph_name]

        el.class_name = " ".join(tokens)
        return el



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Connect REST API with Analyzers runner.
"""
#
# Imports =====================================================================
import traceback
from os.path import join

import requests  # for requests.exceptions.Timeout

from bottle import post
from bottle import response
from bottle_rest import form_to_params

from ..zeo import RequestDatabase
from ..settings import API_PATH
from ..settings import REQUEST_TIMEOUT

from shared import JSON_MIME


# REST API ====================================================================
@post(join(API_PATH, "analyze"))
@form_to_params
def analyzer_api(url):
    """
    Analyze given `url` and return output as JSON.
    """
    rd = RequestDatabase()
    response.content_type = JSON_MIME

    # handle cacheing
    ri = rd.get_request(url)

    try:

        if ri.is_old():
            print "Running the analysis"  #: TODO: to log

            # forget the old one and create new request info - this prevents
            # conflict errors
            ri = rd.get_request(url, new=True)

            # run the processing
            ri.paralel_processing()

    except (requests.exceptions.Timeout, requests.ConnectionError) as e:
        error_msg = """
            Požadovanou stránku {url} nebylo možné stáhnout během {timeout}
            vteřin. Zkuste URL zadat s `www.`, či zkontrolovat funkčnost
            stránek.
            <span style="display: none">{message}</span>
        """
        error_msg = error_msg.format(
            url=url,
            timeout=REQUEST_TIMEOUT,
            message=str(e.message)
        )

        return {
            "status": False,
            "log": ri.get_log(),
            "error": error_msg,
        }
    except Exception as e:
        return {
            "status": False,
            "log": ri.get_log(),
            "error": str(e.message) + "\n" + traceback.format_exc().strip()
        }

    return {
        "status": True,
        "body": ri.to_dict(),
        "log": ri.get_log(),
    }



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from request_database import RequestDatabase



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
"""
Module is containing all necessary global variables for the package.

Module also has the ability to read user-defined data from following paths:

- ``SETTINGS_PATH`` env variable file pointer to .json file.
- ``$HOME/_SETTINGS_PATH``
- ``/etc/_SETTINGS_PATH``

See :attr:`_SETTINGS_PATH` for details.

Note:
    When the first path is found, others is ignored.

Example of the configuration file (``$HOME/webarchive/wa_kat.json``)::

    {
        "WEB_ADDR": "somedomain.cz",
        "WEB_PORT": 80
    }

Example of starting the program with env variable::

    export WA_KAT_SETTINGS="/tmp/conf.json"; bin/wa_kat_server.py

Attributes
----------
"""
# Imports =====================================================================
import os
import json
import os.path


# Functions ===================================================================
def _in_conf_dir(filename):
    """
    Put the `filename` to the configuration directory context / path.
    """
    return os.path.join(
        os.path.dirname(__file__),
        "templates/conf",
        filename
    )


# Module configuration ========================================================
#: Path to the file with zeo_client.conf.
ZEO_CLIENT_PATH = _in_conf_dir("zeo_client.conf")
ZEO_SERVER_PATH = _in_conf_dir("zeo.conf")  #: ZEO server configuration.
PROJECT_KEY = "wa_kat"  #: This is used in ZODB. DON'T CHANGE THIS.
ZEO_CACHE_TIME = 60 * 30  #: ZEO cache time - 30 minutes.
ZEO_MAX_WAIT_TIME = 60 * 5  #: Time after which the processing is restarted. 5m

WEB_ADDR = "0.0.0.0"  #: Address where the webserver should listen.
WEB_PORT = 8080  #: Port for the webserver.
WEB_SERVER = 'paste'  #: Use `paste` for threading.
WEB_DEBUG = False  #: Turn on web debug messages?
WEB_RELOADER = False  #: Turn on reloader for webserver?
WEB_BE_QUIET = False  #: Be quiet and don't emit debug messages to terminal.

#: Settings for the Seeder API
SEEDER_INFO_URL = "http://seeder.visgean.me/api/source/%s/"
SEEDER_TOKEN = ""  #: Token for the Seeder API
SEEDER_TIMEOUT = 5  #: How long to wait for the Seeder.

#: Error message in case that the analysis timeouted.
TIMEOUT_MESSAGE = (
    "Požadovanou stránku nebylo možné stáhnout. "
    "Zkuste zadat URL s www."
)
REQUEST_TIMEOUT = 5  #: How long to wait until the analysed page loads.

GUI_TO_REST_PERIODE = 2  #: How often check the REST API.

WHOIS_URL = "http://whois.icann.org/en/lookup?name=%s"  #: Whois URL
NTK_ALEPH_URL = "http://aleph.techlib.cz/X"  #: URL to the NTK aleph.

#: User agent used in analysis.
USER_AGENT = "http://webarchiv.cz catalogization tool WA-KAT."

API_PATH = "/api_v1/"  #: Path for the REST API.


# User configuration reader (don't edit this) =================================
_ALLOWED = [str, unicode, int, float, long, bool]  #: Allowed types.
_SETTINGS_PATH = "webarchive/wa_kat.json"  #: Appended to default search paths.


def _get_all_constants():
    """
    Get list of all uppercase, non-private globals (doesn't start with ``_``).

    Returns:
        list: Uppercase names defined in `globals()` (variables from this \
              module).
    """
    return [
        key for key in globals().keys()
        if all([
            not key.startswith("_"),          # publicly accesible
            key.upper() == key,               # uppercase
            type(globals()[key]) in _ALLOWED  # and with type from _ALLOWED
        ])
    ]


def _substitute_globals(config_dict):
    """
    Set global variables to values defined in `config_dict`.

    Args:
        config_dict (dict): dict with data, which are used to set `globals`.

    Note:
        `config_dict` have to be dictionary, or it is ignored. Also all
        variables, that are not already in globals, or are not types defined in
        :attr:`_ALLOWED` (str, int, ..) or starts with ``_`` are silently
        ignored.
    """
    if not isinstance(config_dict, dict):
        raise ValueError("Configuration file must be contained in dictionary!")

    constants = _get_all_constants()
    for key, val in config_dict.iteritems():
        if key in constants and type(val) in _ALLOWED:
            globals()[key] = val


def _assert_constraints():
    def _format_error(var_name, msg):
        msg = repr(msg) if msg else "UNSET!"
        return "You have to set %s (%s) in rest.json config!" % (var_name, msg)

    def _assert_var_is_set(var_name):
        assert globals()[var_name], _format_error(
            var_name,
            globals()[var_name]
        )

    def _assert_exists_and_perm(var, path, perm):
        msg = "Can't access the required `%s` " % path
        msg += "file set in configuration (%s)!" % var

        assert os.path.exists(path) and os.access(path, perm), msg

    _assert_var_is_set("ZEO_CLIENT_PATH")
    _assert_exists_and_perm("ZEO_CLIENT_PATH", ZEO_CLIENT_PATH, os.R_OK)


def _read_from_paths():
    """
    Try to read data from configuration paths ($HOME/_SETTINGS_PATH,
    /etc/_SETTINGS_PATH).
    """
    home = os.environ.get("HOME", "")
    home_path = os.path.join(home, _SETTINGS_PATH)
    etc_path = os.path.join("/etc", _SETTINGS_PATH)
    env_path = os.environ.get("SETTINGS_PATH", "")

    read_path = None
    if env_path and os.path.exists(env_path):
        read_path = env_path
    elif home and os.path.exists(home_path):
        read_path = home_path
    elif os.path.exists(etc_path):
        read_path = etc_path

    if not read_path:
        return "{}"

    with open(read_path) as f:
        return f.read()


def _apply_settings():
    """
    Read variables from the possible paths. Assert that constraints are set.
    """
    _substitute_globals(
        json.loads(_read_from_paths())
    )

    if not os.environ.get('READTHEDOCS'):
        _assert_constraints()


_apply_settings()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This file provides virtual fileystem for Brython files, so some of the values
may be transported to the frontend from backend as pre-parsed python data.
"""
#
# Imports =====================================================================
import json
from os.path import join
from functools import wraps

from bottle import get
from bottle import response
from backports.functools_lru_cache import lru_cache

from .. import settings

from shared import read_template


# Variables ===================================================================
#: This is used as header of all virtual .py files
PY_HEADER = """#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
"""


# Functions & classes =========================================================
def in_virtual_path(fn):
    """
    Return `fn` in absolute path to root of the brython files.

    Args:
        fn (str): Name of the file which should be put into abs path.

    Returns:
        str: Absolute path of the file.
    """
    return join("/static/js/Lib/site-packages/virtual_fs", fn)


def in_second_virtual_path(fn):
    """
    Brython has problem with .. imports, so I have to insert the `submodule`
    package into the root path manually.

    Args:
        fn (str): Name of the file which should be put into abs path.

    Returns:
        str: Absolute path of the file in `components` submodule.
    """
    return join("/static/js/Lib/site-packages/components/virtual_fs", fn)


def python_mime(fn):
    """
    Decorator, which adds correct MIME type for python source to the decorated
    bottle API function.
    """
    @wraps(fn)
    def python_mime_decorator(*args, **kwargs):
        response.content_type = "text/x-python"

        return fn(*args, **kwargs)

    return python_mime_decorator


# API =========================================================================
@get(in_virtual_path("__init__.py"))
@get(in_second_virtual_path("__init__.py"))
@python_mime
def init_api():
    """
    Virtual ``__init__.py`` file for the whole ``virtual_fs/`` directory.
    """
    return PY_HEADER


@get(in_virtual_path("settings.py"))
@get(in_second_virtual_path("settings.py"))
@python_mime
@lru_cache()
def settings_api():
    """
    Virtual ``settings.py`` with values transported from real
    :mod:`settings.py`, so the Brython frontend may be configured same way as
    backend.

    Some of the critical values are intentionally left out.
    """
    variables = [
        "%s = %s" % (var, repr(getattr(settings, var)))
        for var in sorted(dir(settings))
        if (not var.startswith("_") and var.upper() == var and
            not var.startswith("SEEDER"))  # hide the private tokens and url
    ]

    return PY_HEADER + "\n\n".join(variables)


@get(in_virtual_path("conspectus.py"))
@get(in_second_virtual_path("conspectus.py"))
@python_mime
@lru_cache()
def conspectus_api():
    """
    Virtual ``conspectus.py`` file for brython. This file contains following
    variables:

    Attributes:
        consp_dict (dict): Dictionary with conspects.
        cosp_id_pairs (list): List of tuples ``(name, id)`` for conspectus.
        subs_by_mdt (dict): Dictionary containing ``mdt: sub_dict``.
        mdt_by_name (dict): ``name: mdt`` mapping.

    Note:
        Example of the `cons_dict` format::

            {
                "1": {
                    "id": "1",
                    "name": "Antropologie, etnografie",
                    "subconspects": {
                        "304": {
                            "conspect_id": "1",
                            "name": "Kulturn\u00ed politika"
                            "en_name": "Culture and institutions",
                            "mdt": "304",
                            "ddc": "306.2",
                        },
                        ...
                    }
                },
                ...
            }

    Values are stored in json and upacked after load. This is used because of
    performance issues with parsing large brython files.
    """
    def to_json(data):
        """
        JSON conversion is used, because brython has BIG performance issues
        when parsing large python sources. It is actually cheaper to just load
        it as JSON objects.

        Load times:
            Brython: 24s firefox, 54s chromium
            JSON: 14s firefox, 9s chromium
        """
        return repr(json.dumps(data))

    conspectus_dict = json.loads(read_template("full_conspectus.json"))

    # raw conspectus dictionary
    out = PY_HEADER
    out += "import json\n"

    out += "consp_dict = json.loads(%s)\n\n" % to_json(conspectus_dict)

    # (consp, id) pairs
    cosp_id_pairs = sorted(
        (x["name"], x["id"])
        for x in conspectus_dict.values()
    )
    out += "cosp_id_pairs = json.loads(%s)\n\n" % to_json(cosp_id_pairs)

    # mdt -> subconspect mapping
    subs_by_mdt = [
        x["subconspects"].values()
        for x in conspectus_dict.values()
    ]
    subs_by_mdt = {d["mdt"]: d for d in sum(subs_by_mdt, [])}
    out += "subs_by_mdt = json.loads(%s)\n\n" % to_json(subs_by_mdt)

    # subconspect_name -> subconspect mapping
    out += "mdt_by_name = json.loads(%s)\n\n" % to_json({
        d["name"]: d["mdt"]
        for d in subs_by_mdt.values()
    })

    return out


@get(in_virtual_path("periodes.py"))
@get(in_second_virtual_path("periodes.py"))
@python_mime
@lru_cache()
def periode_api():
    """
    Virtual ``periodes.py`` file with list of possible periodes values in
    variable ``periode_list``.
    """
    periodes = read_template("periode.txt").decode("utf-8")

    return PY_HEADER + "periode_list = %s.splitlines()\n\n" % repr(periodes)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Functions for parsing informations about date when the page / domain was
created.
"""
#
# Imports =====================================================================
import requests
import pythonwhois
from kwargs_obj import KwargsObj

from ..settings import WHOIS_URL


# Variables ===================================================================
class TimeResource(KwargsObj):
    """
    This class is backward compatible with :class:`.SourceString`, so it may
    be used to transport backend data to frontend.

    Attributes:
        url (str): URL of the additional informations.
        date (str): ISO 8601 date strings.
        val (str): Year as string.
        source (str): Information about origin of the source.
    """
    def __init__(self, **kwargs):
        self.url = None
        self.date = None

        # for backward compatibility with SourceString
        self.val = None
        self.source = None

        self._kwargs_to_attributes(kwargs)

    def to_dict(self):
        """
        Convert yourself to dictionary.
        """
        return {
            key: val
            for key, val in self.__dict__.iteritems()
        }

    def __str__(self):
        return self.val

    def __repr__(self):
        return "%s(%s)" % (self.source, self.val)


# Functions & classes =========================================================
def mementoweb_api_tags(url):
    """
    Parse list of :class:`TimeResource` objects based on the mementoweb.org.

    Args:
        url (str): Any url.

    Returns:
        list: :class:`TimeResource` objects.
    """
    memento_url = "http://labs.mementoweb.org/timemap/json/"

    r = requests.get(memento_url + url)

    if r.status_code != 200:
        return []

    data = r.json().get("mementos", {}).get("list", [])

    if not data:
        return []

    resources = (
        TimeResource(
            url=item.get("uri", ""),
            date=item.get("datetime", ""),
            val=item.get("datetime", "").split("-")[0],
            source="MementoWeb.org",
        )
        for item in data
    )

    # deduplicate the resources
    resource_dict = {
        res.val: res
        for res in resources
    }

    return resource_dict.values()


def get_whois_tags(domain):
    """
    Get :class:`TimeResource` objects with creation dates from Whois database.

    Args:
        domain (str): Domain without http://, relative paths and so on.

    Returns:
        list: :class:`TimeResource` objects.
    """
    data = pythonwhois.get_whois("kitakitsune.org")

    return [
        TimeResource(
            url=WHOIS_URL % domain.strip(),
            date=date.isoformat("T"),
            val=date.strftime("%Y"),
            source="Whois",
        )
        for date in data.get("creation_date", [])
    ]


def get_creation_date_tags(url, domain, as_dicts=False):
    """
    Put together all data sources in this module and return it's output.

    Args:
        url (str): URL of the web. With relative paths and so on.
        domain (str): Just the domain of the web.
        as_dicts (bool, default False): Convert output to dictionaries
            compatible with :class:`.SourceString`?

    Returns:
        list: Sorted list of :class:`TimeResource` objects or dicts.
    """
    creation_date_tags = [
        mementoweb_api_tags(url),
        get_whois_tags(domain),
    ]

    creation_date_tags = sorted(
        sum(creation_date_tags, []),
        key=lambda x: x.date
    )

    if not as_dicts:
        return creation_date_tags

    return [
        item._as_dict()
        for item in creation_date_tags
    ]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from os.path import join

from browser import window
from browser import document

# virtual filesystem / modules provided by REST API
from virtual_fs import settings


# Functions & classes =========================================================
class KeywordListHandler(object):
    """
    This class is used to control the GUI for the list of keywords.

    It allows user to add new keyword, remove present keyword and get a list
    of defined keywords.
    """
    def __init__(self, el_id, whole_id=None):
        if whole_id is None:
            whole_id = "whole_" + el_id

        self.el = document[el_id]
        self.whole_el = document[whole_id]
        self.all_lists_el = document["whole_keyword_list"]

        self.keywords = []

        self._remover = """
            <span class='kw_remover'
                  title='Odstranit klíčové slovo.'
                  id='kw_remover_id_%d'>
                ✖
            </span>
        """

        self._render()

    def _render(self):
        """
        Render the HTML code for all the :attr:`keywords` stored in this class.

        This method is called after each change in :attr:`keywords`.
        """
        # hide the list in case that there is no `keyword` to be displayed
        if self.keywords:
            self.whole_el.style.display = "block"
            self.all_lists_el.style.display = "block"
        else:
            self.whole_el.style.display = "none"
            if "<li>" not in self.all_lists_el.html:
                self.all_lists_el.style.display = "none"

        # construct the HTML code for each keyword
        html_lines = (
            "<li class='kw_enum'>{0} {1}</li>\n".format(
                keyword,
                (self._remover % cnt)
            )
            for cnt, keyword in enumerate(self.keywords)
        )

        # put the keywords into the HTML code of the page
        self.el.innerHTML = "<ol>\n%s\n</ol>\n" % "\n".join(html_lines)

        # this function is used to bind the ✖ to function for removing the
        # keyword
        def keyword_remover(keyword):
            def remover(ev):
                self.remove_keyword(keyword)

            return remover

        # go thru all the keywords and bind them to keyword_remover()
        for cnt, keyword in enumerate(self.keywords):
            uid = "kw_remover_id_%d" % cnt
            el = document[uid]
            el.bind("click", keyword_remover(keyword))

    def add_keyword(self, keyword):
        """
        Add `keyword` to :attr:`keywords`.

        Args:
            keyword (str): New keyword.
        """
        self.keywords.append(keyword)
        self._render()

    def remove_keyword(self, keyword):
        """
        Remove `keyword` from :attr:`keywords`.

        Args:
            keyword (str): Keyword which should be removed.
        """
        self.keywords.remove(keyword)
        self._render()

    def reset(self):
        """
        Reset the widget to the default state.
        """
        self.keywords = []
        self._render()


UserKeywordHandler = KeywordListHandler("user_keyword_list")
AlephKeywordHandler = KeywordListHandler("aleph_keyword_list")
AanalysisKeywordHandler = KeywordListHandler("analysis_keyword_list")


class KeywordAdder(object):
    """
    This class is here to controll typeahead input bar with keyword suggestion.

    Keywords selected from suggestions are then mapped to
    :class:`KeywordListHandler`.
    """
    intput_el = document["keyword_adder"]

    @classmethod
    def on_select_callback(cls, selected_item):
        """
        This method defines the action taken when user selects the keyword from
        suggestion engine.

        Args:
            selected_item (str): Keyword selected by the user.

        Returns:
            str: Value on which the <input> element will be set.
        """
        UserKeywordHandler.add_keyword(selected_item)

        return ""

    @classmethod
    def set_kw_typeahead_input(cls):
        """
        Map the typeahead input to remote dataset.
        """
        # get reference to parent element
        parent_id = cls.intput_el.parent.id
        if "typeahead" not in parent_id.lower():
            parent_id = cls.intput_el.parent.parent.id

        window.make_keyword_typeahead_tag(
            "#" + parent_id,
            join(settings.API_PATH, "kw_list.json"),
            cls.on_select_callback,
        )

KeywordAdder.set_kw_typeahead_input()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from browser import html
from browser import alert
from browser import window
from browser import document

from virtual_fs import conspectus


# Functions & classes =========================================================
class ConspectHandler(object):
    """
    Conspect is this dataset, where there is 21 categories (conspects) and
    hundreds of sub-categories (sub-conspects).

    This class allow user to pick the conspect / subconspect elements using
    following widgets:

        - two selects
        - one searchable typeahead tag

    First approach may be better when you don't exactly know what do you want
    and feel like browsing the conspects.

    Second approach is better when you just simply need to pick one.

    User may switch between this two approaches without the loss of the
    information.
    """
    input_el = document["conspect_subconspect"]
    conspect_el = document["konspekt"]
    subconspect_el = document["subkonspekt"]

    is_twoconspect = True

    @classmethod
    def _save_value(cls, value):
        """
        Callback used to automatically store value in :attr:`value` property
        each time the user selects value.
        """
        cls.value = value

    @classmethod
    def _get_subconspects(cls, consp_id):
        subconspects = conspectus.consp_dict[consp_id]["subconspects"].values()

        return sorted(subconspects, key=lambda x: x["name"])

    @classmethod
    def _draw_subconspects(cls, consp_id):
        cls.subconspect_el.html = ""
        for sub_conspect in cls._get_subconspects(consp_id):
            name = sub_conspect["name"]
            mdt = sub_conspect["mdt"]

            cls.subconspect_el <= html.OPTION(name, value=mdt)

    @classmethod
    def _draw_conspects(cls):
        for name, uid in conspectus.cosp_id_pairs:
            cls.conspect_el <= html.OPTION(name, value=uid)

        cls.conspect_el.bind(
            'change',
            lambda ev: cls._draw_subconspects(ev.target.value)
        )

    @classmethod
    def _create_searchable_typeahead(cls):
        """
        Create typeahead <input> element and fill it with data.
        """
        window.destroy_typeahead_tag("#conspect_subconspect_typeahead")
        args = [
            {
                "name": name,
                "data": [x["name"] for x in cls._get_subconspects(uid)],
            }
            for name, uid in conspectus.cosp_id_pairs
        ]
        window.make_multi_searchable_typeahead_tag(
            "#conspect_subconspect_typeahead",
            *args
        )

    @classmethod
    def get(cls):
        """
        Get code selected by user.

        Returns:
            str: Code or None in case that user didn't selected anything yet.
        """
        if cls.is_twoconspect:
            return cls.subconspect_el.value or None

        input_value = cls.input_el.value.strip()

        # blank user input -> no value was yet set
        if not input_value:
            return None

        mdt = conspectus.mdt_by_name.get(input_value)

        if not mdt:
            alert("Invalid sub-conspect `%s`!" % input_value)
            return None

        return mdt

    @classmethod
    def set(cls, mdt):
        """
        Set value for <input> / <select> tags based on MDT.

        Args:
            mdt (str): MDT of the conspect / subconspect.
        """
        if type(mdt) in [list, tuple]:
            mdt = mdt[0]

        if isinstance(mdt, dict):
            mdt = mdt["val"]

        if not mdt:
            cls.conspect_el.value = ""
            cls.subconspect_el.html = ""
            cls.input_el.value = ""
            return

        # set twoconspect
        data = conspectus.subs_by_mdt.get(mdt)

        if data:
            cls.conspect_el.value = data["conspect_id"]

            cls._draw_subconspects(data["conspect_id"])
            cls.subconspect_el.value = mdt

            # set input
            cls.input_el.value = data["name"]
        else:
            cls.input_el.value = mdt + " not found"

    @classmethod
    def get_dict(cls):
        """
        Return dictionary with conspect / subconspect info.
        """
        mdt = cls.get()

        if not mdt:
            return {}

        return conspectus.subs_by_mdt.get(mdt, {})


class ConspectDescriptor(ConspectHandler):
    switcher_els = document.get(selector=".conspect_switcher")
    two_conspect_el = document.get(selector=".two_conspect")
    whole_conspect_subconspect_el = document["whole_conspect_subconspect"]

    @classmethod
    def bind_switcher(cls):
        """
        Bind the switch checkbox to functions for switching between types of
        inputs.
        """
        def show_two_conspect():
            cls.is_twoconspect = True

            # search by class
            for el in cls.two_conspect_el:
                el.style.display = "block"

            cls.whole_conspect_subconspect_el.style.display = "none"

        def hide_two_conspect():
            cls.is_twoconspect = False

            for el in cls.two_conspect_el:
                el.style.display = "none"

            cls.whole_conspect_subconspect_el.style.display = "block"

        def show_or_hide_two_conspect(ev):
            val = cls.get()

            # check / uncheck both checkboxes
            for el in cls.switcher_els:
                el.checked = ev.target.checked

            if ev.target.checked:
                hide_two_conspect()
                cls.set(val)
                return

            show_two_conspect()
            cls.set(val)

        # bind both `conspect_switchers`
        for el in document.get(selector=".conspect_switcher"):
            el.bind("change", show_or_hide_two_conspect)

    @classmethod
    def show_error(cls, error=True):
        """
        Show `error` around the conspect elements. If the `error` is ``False``,
        hide it.
        """
        if error:
            cls.input_el.style.border = "2px solid red"
            cls.conspect_el.style.border = "2px solid red"
            cls.subconspect_el.style.border = "2px solid red"
        else:
            cls.input_el.style.border = "0"
            cls.conspect_el.style.border = "0"
            cls.subconspect_el.style.border = "0"

    @classmethod
    def init(cls):
        """
        Bind elements to callbacks.
        """
        for el in cls.switcher_els:
            el.checked = False

        cls.bind_switcher()

        cls._draw_conspects()
        cls._create_searchable_typeahead()

    @classmethod
    def __get__(cls, obj, objtype):
        return cls.get_dict()

    @classmethod
    def __set__(cls, obj, value):
        cls.show_error(False)
        cls.set(value)

    @classmethod
    def validate(cls):
        """
        Make sure, that conspect element is properly selected. If not, show
        error.
        """
        if cls.get_dict():
            cls.show_error(False)
            return True

        cls.show_error(True)
        return False

    @classmethod
    def reset(cls):
        """
        Reset the conspect elements to initial state.
        """
        cls.input_el.value = ""
        cls.subconspect_el.html = ""
        cls.show_error(False)


ConspectDescriptor.init()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from title_detector import get_title_tags
from place_detector import get_place_tags
from author_detector import get_author_tags
from language_detector import get_lang_tags
from keyword_detector import get_keyword_tags
from annotation_detector import get_annotation_tags
from creation_date_detector import get_creation_date_tags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
Generic and specific Error box handlers.
"""
#
# Imports =====================================================================
from browser import document


# Functions & classes =========================================================
class ErrorBox(object):
    """
    Generic class for showing the errors in nice red box.
    """
    def __init__(self, tag_id, whole_tag_id):
        self.tag = document[tag_id]
        self.whole_tag = document[whole_tag_id]

    def show(self, msg):
        self.tag.innerHTML = msg
        self.whole_tag.style.display = "block"

    def hide(self):
        self.whole_tag.style.display = "none"

    def reset(self):
        self.hide()
        self.tag.innerHTML = ""


UrlBoxError = ErrorBox("urlbox_error", "whole_urlbox_error")
ISSNBoxError = ErrorBox("issnbox_error", "whole_issnbox_error")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import time

import pytest
import transaction

from wa_kat.zeo import RequestDatabase
from wa_kat.zeo.worker import worker
from wa_kat.zeo.request_info import worker_mapping


# Variables ===================================================================
TEST_URL = "http://kitakitsune.org"


# Fixtures ====================================================================
@pytest.fixture
def rdb(client_conf_path):
    return RequestDatabase(conf_path=client_conf_path)


@pytest.fixture
def request_info(client_conf_path):
    # make sure, that new object is created and thus it is indeed new
    # connection to database
    return rdb(client_conf_path).get_request(TEST_URL)


# Functions ===================================================================
def circuit_breaker_progress_retry(ri, attempts=20):
    for i in range(attempts):
        try:
            print ri.progress()
            assert ri.is_all_set()
            return
        except AssertionError:
            time.sleep(1)

    raise IOError("Paralel processing took too much time, aborting.")


# Tests =======================================================================
def test_request_database(rdb):
    assert rdb

    # with transaction.manager:
        # assert not rdb.requests

    with pytest.raises(ValueError):
        rdb.get_request("azgabas")

    assert rdb.get_request(TEST_URL)


def test_worker(rdb, request_info, client_conf_path):
    assert request_info.title_tags is None
    request_info.index = """<HTML>
<head>
    <title>Bystroushaak</title>
    </head>
</HTML>
    """

    property_info = worker_mapping()["title_tags"]
    worker(
        url_key=request_info.url,
        property_info=property_info,
        filler_params=property_info.filler_params(request_info),
        conf_path=client_conf_path,  # to monkeypatch the zeo client conf path
    )

    # lookup to the request info properties
    with transaction.manager:
        assert request_info.title_tags
        assert isinstance(request_info.title_tags, list)

    # lookup to the database
    request = rdb.get_request(request_info.url)
    assert request.title_tags
    assert isinstance(request.title_tags, list)


def test_RequestInfo_paralel_download(request_info, client_conf_path):
    with transaction.manager:
        assert request_info.url == TEST_URL
        assert not request_info.index
        assert not request_info.is_all_set()
        assert request_info.progress()[0] == 1  # set from test_worker()

    request_info.paralel_processing(client_conf=client_conf_path)

    # wait until the processing of all requests is finished
    circuit_breaker_progress_retry(request_info, attempts=20)

    with transaction.manager:
        assert request_info.url
        assert request_info.domain
        assert request_info.index
        assert request_info.creation_ts
        assert request_info.downloaded_ts
        assert request_info.processing_started_ts
        assert request_info.processing_ended_ts

        assert request_info.is_all_set()
        assert request_info.progress()[0] == request_info.progress()[1]

        set_keys = set(request_info.to_dict()["values"].keys())
        assert set_keys == set(worker_mapping().keys())

        assert request_info.to_dict()["values"]["place_tags"]


def test_garbage_collection(rdb):
    with transaction.manager:
        assert rdb.requests

    time.sleep(1)
    rdb.garbage_collection(1)

    with transaction.manager:
        assert not rdb.requests



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This sub-module is used to convert data from the WA-KAT / user inputs to
`Dublin core <http://dublincore.org>`_.
"""
#
# Imports =====================================================================
from collections import OrderedDict

from xmltodict import unparse
from odictliteral import odict


# Functions & classes =========================================================
def _convert_metadata(data):
    """
    Convert metadata from WA-KAT to Dublin core dictionary like structure,
    which may be easily converted to xml using :mod:`xmltodict` module.

    Args:
        data (dict): Nested WA-KAT data. See tests for example.

    Returns:
        dict: Dict in dublin core format.
    """
    def compose(val, arguments=None):
        """
        If not val, return None.

        If not arguments, return just val.

        Otherwise, map val into `arguments` dict under the ``#text`` key.

        This is format required by xmltodict.
        """
        if val is None:
            return None

        if not arguments:
            return val

        arguments["#text"] = val
        return arguments

    conspect = data.get("conspect", {})
    author_name = data.get("author", {}).get("name")
    author_code = data.get("author", {}).get("code")

    metadata = odict[
        "dc:title": data.get("title"),
        "dcterms:alternative": data.get("subtitle"),
        "dc:creator": compose(author_name, {"@id": author_code}),
        "dc:publisher": data.get("publisher"),
        "dc:description": data.get("annotation"),
        "dc:coverage": compose(data.get("place"), {"@xml:lang": "cze"}),
        "dc:language": compose(data.get("language"), {"@schema": "ISO 639-2"}),
        "dcterms:created": data.get("from_year"),
        "dcterms:accrualperiodicity": compose(
            data.get("periodicity"),
            {"@xml:lang": "cze"}
        ),
        "dc:identifier": [
            {"@rdf:resource": data["url"]},
            compose(data.get("issn"), {"@xsi:type": "ISSN"}),
            compose(conspect.get("mdt"), {"@xsi:type": "MDT"}),
            compose(conspect.get("ddc"), {"@xsi:type": "DDC"}),
        ],
        "dc:subject": [
            compose(conspect.get("mdt"), {"@xsi:type": "dcterms:UDC"}),
            compose(conspect.get("ddc"), {"@xsi:type": "dcterms:DDC"}),
        ],
    ]

    def pick_keywords(data, source):
        """
        Convert::

            "en_keywords": [
                {
                    "zahlavi": "keyboard (musical instrument)",
                    "zdroj": "eczenas"
                }, {
                    "zahlavi": "ANCA-associated vasculitis",
                    "zdroj": "eczenas"
                }
            ]

        To::
            ["keyboard (musical instrument)", "ANCA-associated vasculitis"]
        """
        return [
            x["zahlavi"]
            for x in data.get(source, [])
            if x.get("zahlavi")
        ]

    # parse and add keywords (keywords are in dicts with other data, I want
    # just the free-text descriptions)
    cz_keywords = pick_keywords(data, "cz_keywords")
    en_keywords = pick_keywords(data, "en_keywords")

    if cz_keywords:
        metadata["dc:subject"].append({
            "@xml:lang": "cz",
            "#text": ", ".join(cz_keywords)
        })
    if en_keywords:
        metadata["dc:subject"].append({
            "@xml:lang": "en",
            "#text": ", ".join(en_keywords)
        })

    # filter unset identifiers - TODO: rewrite to recursive alg.
    metadata["dc:identifier"] = [x for x in metadata["dc:identifier"] if x]
    metadata["dc:subject"] = [x for x in metadata["dc:subject"] if x]

    return metadata


def _remove_none(data):
    """
    Go thru `data` (dict / list) and remove ``None`` values.
    """
    if isinstance(data, list) or isinstance(data, tuple):
        return [x for x in data if x is not None]

    out = OrderedDict()
    for key, val in data.iteritems():
        if val is not None:
            out[key] = val

    return out


def to_dc(data):
    """
    Convert WA-KAT `data` to Dublin core XML.

    Args:
        data (dict): Nested WA-KAT data. See tests for example.

    Returns:
        unicode: XML with dublin core.
    """
    root = odict[
        "metadata": odict[
            "@xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "@xmlns:dc": "http://purl.org/dc/elements/1.1/",
            "@xmlns:dcterms": "http://purl.org/dc/terms/",
            "@xmlns:rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        ]
    ]

    # map metadata to the root element, skip None values
    for key, val in _convert_metadata(_remove_none(data)).iteritems():
        if val is None:
            continue

        if isinstance(val, basestring) and not val.strip():
            continue

        if isinstance(val, str):
            val = val.decode("utf-8")

        root["metadata"][key] = val

    return unparse(root, pretty=True, indent="    ")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from wa_kat.analyzers.language_detector import get_lang_tags


# Variables ===================================================================
TEST_TEMPLATE = """
<HTML>
<head>
    <title>HTML title</title>

    <meta name    = "DC.Language"
          content = "cs">

    <meta http-equiv="Content-language"
          content="cs">
</head>
<body>
Zde je nějaký ten obsah, který by měl být rozpoznaný jako čeština.
</body>
</HTML>
"""


# Tests =======================================================================
def test_get_lang_tags():
    lang_tags = get_lang_tags(TEST_TEMPLATE)

    assert lang_tags[0] == "cze"
    assert lang_tags[0].source == "HTML"

    assert lang_tags[1] == "cze"
    assert lang_tags[1].source == "DC"

    assert lang_tags[2] == "cze"
    assert lang_tags[2].source == "langdetect"



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from browser import ajax


# Functions & classes =========================================================
def make_request(url, data, on_complete):
    """
    Make AJAX request to `url` with given POST `data`. Call `on_complete`
    callback when complete.

    Args:
        url (str): URL.
        data (dict): Dictionary with POST data.
        on_complete (ref): Reference to function / method which will be called
            when the request is done.
    """
    req = ajax.ajax()
    req.bind('complete', on_complete)
    req.open('POST', url, True)
    req.set_header('content-type', 'application/x-www-form-urlencoded')
    req.send(data)


def func_on_enter(func):
    """
    Register the `func` as a callback reacting only to ENTER.

    Note:
        This function doesn't bind the key to the element, just creates sort of
        filter, which ignores all other events.
    """
    def function_after_enter_pressed(ev):
        ev.stopPropagation()

        # if the key was `enter` ..
        if ev.keyCode == 13:
            func(ev)

    return function_after_enter_pressed



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
Interface which handles the show/hide button for Author picker.
"""
#
# Imports =====================================================================
from browser import document


# Variables ===================================================================
AUTHOR_EL = document["whole_author"]
SWITCHER_EL = document["author_switcher"]


# Functions & classes =========================================================
def show_author(ev=None):
    AUTHOR_EL.style.display = "block"

    SWITCHER_EL.checked = True
    SWITCHER_EL.bind("change", hide_author)


def hide_author(ev=None):
    AUTHOR_EL.style.display = "none"

    SWITCHER_EL.checked = False
    SWITCHER_EL.bind("change", show_author)


hide_author()
SWITCHER_EL.bind("change", show_author)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
This file is used as definition of data:

    - stored in database
    - sent from backend to frontend
    - transported from Aleph to frontend
"""
#
# Imports =====================================================================
from collections import namedtuple

from kwargs_obj import KwargsObj

import analyzers


# Functions & classes =========================================================
class FuncInfo(namedtuple("FuncInfo", ["func", "args_func"])):
    """
    Container for functions and it's arguments.

    Attributes:
        func (ref): Function reference.
        args_func (list): List of function arguments.
    """


def _compose_func(func, args_func=lambda req_info: [req_info.index]):
    """
    Compose function used to compose arguments to function.

    Arguments for the functions are composed from the :class:`.RequestInfo`
    object from the ZODB.
    """
    return FuncInfo(func=func, args_func=args_func)


class Model(KwargsObj):
    """
    This class is used as definition of all fields transported from backend
    to frontend (analysis, Aleph query), and also all fields stored in
    database.

    Any of the field may be set to ``None`` and in that case is ignore.
    """
    def __init__(self, **kwargs):
        self.url = None  #: URL of the resource.
        self.title_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.subtitle_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.creation_dates = None  #: List of dicts ``{source: .., val: ..}``.
        self.author_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.publisher_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.place_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.keyword_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.conspect = None  #: Conspect code.
        self.lang_tags = None  #: List of dicts ``{source: .., val: ..}``.
        self.annotation_tags = None  #: List of dicts ``{source: .., val:..}``.
        self.periodicity = None  #: Periodicity string.
        self.source_info = None  #: 500a field from Aleph.
        self.original_xml = None  #: MARC XML record from Aleph.
        self.issn = None  #: ISSN, if defined.
        self.rules = None  #: Rules from the Seeder.
        self.additional_info = None  #: Additional fields from Aleph.

        self._kwargs_to_attributes(kwargs)

    @classmethod
    def analyzers_mapping(cls):
        """
        Return instance of itself where all used properties are set to
        :class:`FuncInfo`.

        This method is used by the database, which map all the properties
        defined here to itself, runs the functions as new processes and stores
        the result in itself. Because it knows how many properties are there,
        it may also track the progress, which is then transported to the
        frontend and displayed in form of progress bar.

        Returns:
            obj: :class:`Model` instance.
        """
        return cls(
            title_tags=_compose_func(analyzers.get_title_tags),
            place_tags=_compose_func(
                analyzers.get_place_tags,
                lambda req_info: (req_info.index, req_info.domain)
            ),
            lang_tags=_compose_func(analyzers.get_lang_tags),
            keyword_tags=_compose_func(analyzers.get_keyword_tags),
            # yep, authors of webpage are actually publishers
            publisher_tags=_compose_func(analyzers.get_author_tags),
            annotation_tags=_compose_func(analyzers.get_annotation_tags),
            creation_dates=_compose_func(
                analyzers.get_creation_date_tags,
                lambda req_info: (req_info.url, req_info.domain)
            ),
        )

    def keys(self):
        """
        Return list of properties.
        """
        return self.__dict__.keys()

    def get_mapping(self):  # TODO: rename to _as_dict
        """
        Convert the class to dict.

        Returns:
            dict: Copy of ``self.__dict__``.
        """
        return {
            key: val
            for key, val in self.__dict__.iteritems()
            if val
        }

    def __repr__(self):
        params = ", ".join(
            "%s=%s" % (key, repr(val))
            for key, val in self.__dict__.iteritems()
        )

        return "%s(%s)" % (self.__class__.__name__, params)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
import time
from urlparse import urlparse
from collections import namedtuple
from multiprocessing import Process
from functools import total_ordering

import requests
import transaction
from persistent import Persistent
from zeo_connector import transaction_manager
from backports.functools_lru_cache import lru_cache

from ..settings import USER_AGENT
from ..settings import ZEO_CACHE_TIME
from ..settings import REQUEST_TIMEOUT
from ..settings import ZEO_MAX_WAIT_TIME
from ..data_model import Model

from worker import worker


# Functions & classes =========================================================
@lru_cache()
def worker_mapping():
    return Model().analyzers_mapping().get_mapping()


class Progress(namedtuple("Progress", ["done", "base"])):
    """
    Progress bar representation.

    Attr:
        done (int): How much is done.
        base (int): How much is there.
    """


@total_ordering
class RequestInfo(Persistent):
    """
    This object is used to hold informations about requests, which are
    processed at that moment. It stores all its properites in ZODB.

    The object is basically used as cache AND progress bar indicator of work.

    Attributes:
        url (str): URL to which this info object is related.
        domain (str): Domain of the URL.
        index (str): Content of the index page of URL.
        creation_ts (float): Timestamp of object creation.
        downloaded_ts (float): Timestamp showing where the :attr:`index` was
            downloaded.
        processing_started_ts (float): Timestamp showing when the processing of
            additional properties started.
        processing_ended_ts (float): Timestamp showing when the processing of
            additional properties stopped.

    Warning:
        There is more properties, one for all analyzator function from
        :mod:`.analyzers`.
    """
    def __init__(self, url):
        """
        Constructor.

        Args:
            url (str): URL to which this request is related.
        """
        self.url = url
        self.domain = urlparse(url).netloc
        self.index = None
        self._log = {}

        # time trackers
        self.creation_ts = time.time()
        self.downloaded_ts = None
        self.processing_started_ts = None
        self.processing_ended_ts = None

        for key in worker_mapping().keys():
            setattr(self, key, None)

    def _download(self, url):
        """
        Download `url` and return it as utf-8 encoded text.

        Args:
            url (str): What should be downloaded?

        Returns:
            str: Content of the page.
        """
        headers = {"User-Agent": USER_AGENT}
        resp = requests.get(
            url,
            timeout=REQUEST_TIMEOUT,
            headers=headers,
            allow_redirects=True,
            verify=False,
        )

        def decode(st, alt_encoding=None):
            encodings = ['ascii', 'utf-8', 'iso-8859-1', 'iso-8859-15']

            if alt_encoding:
                if isinstance(alt_encoding, basestring):
                    encodings.append(alt_encoding)
                else:
                    encodings.extend(alt_encoding)

            for encoding in encodings:
                try:
                    return st.encode(encoding).decode("utf-8")
                except UnicodeEncodeError, UnicodeDecodeError:
                    pass

            raise UnicodeError('Could not find encoding.')

        return decode(resp.text, resp.encoding)

    @transaction_manager
    def log(self, msg):
        self._log[time.time()] = msg

    @transaction_manager
    def get_log(self):
        # create banlist
        banlist = set(worker_mapping().keys())
        banlist.add("index")

        # log the state of properties in RequestInfo, except thos in banlist
        property_status = [
            "   %s=%s," % (key, repr(val))
            for key, val in self.__dict__.items()
            if not key.startswith("_") and key not in banlist
        ]

        # construct the namedtuple-like view on data
        status = "%s(\n%s\n)\n---\n" % (
            self.__class__.__name__,
            "\n".join(property_status),
        )

        # add logged messages
        return status + "\n".join(
            "%s: %s" % (str(timestamp), self._log[timestamp])
            for timestamp in sorted(self._log.keys())
        )

    def paralel_processing(self, client_conf=None):
        """
        Lauch paralel processes (see :func:`.worker`) to fill properties with
        data.

        Args:
            client_conf (str, default None): Optional parameter used by tests
                to redirect the database connections to test's environment.
        """
        self._reset_set_properties()

        with transaction.manager:
            self.index = self._download(self.url)
            self.downloaded_ts = time.time()
            self.processing_started_ts = time.time()

        # launch all workers as paralel processes
        for name, function_info in worker_mapping().iteritems():
            p = Process(
                target=worker,
                kwargs={
                    "url_key": self.url,
                    "property_name": name,
                    "function": function_info.func,
                    "function_arguments": function_info.args_func(self),
                    "conf_path": client_conf,
                }
            )
            p.start()

    @transaction_manager
    def _reset_set_properties(self):
        """
        Reset the progress counter back to zero.
        """
        for property_name in worker_mapping().keys():
            setattr(self, property_name, None)

    @transaction_manager
    def _get_all_set_properties(self):
        """
        Collect names of set properties.

        Returns:
            set: Set containing names of all properties, which are set to \
                 non-None value.
        """
        return set(
            property_name
            for property_name in worker_mapping().keys()
            if getattr(self, property_name) is not None
        )

    def progress(self):
        """
        Get progress.

        Returns:
            namedtuple: :class:`Progress`.
        """
        return Progress(
            done=len(self._get_all_set_properties()),
            base=len(worker_mapping()),
        )

    def is_all_set(self):
        """
        Is all properties set?

        Returns:
            bool: True if they are set.
        """
        return self._get_all_set_properties() == set(worker_mapping().keys())

    @transaction_manager
    def is_old(self):
        """
        Is the object cached for too long, so it should be redownloaded?

        See :attr:`.ZEO_MAX_WAIT_TIME` and :attr:`.ZEO_CACHE_TIME` for details.

        Returns:
            bool: True if it is.
        """
        if not self.processing_started_ts:
            return True

        if self.processing_ended_ts:
            return self.processing_ended_ts + ZEO_CACHE_TIME < time.time()

        # in case that processing started, but didn't ended in
        # ZEO_MAX_WAIT_TIME
        expected_end_ts = self.creation_ts + ZEO_MAX_WAIT_TIME
        if expected_end_ts < time.time():
            self.log("Prosessing timeouted and properites were not set!")

        return expected_end_ts < time.time()

    @transaction_manager
    def to_dict(self):
        """
        This method is used in with connection to REST API. It basically
        converts all important properties to dictionary, which may be used by
        frontend.

        Returns:
            dict: ``{"all_set": bool, "progress": [int(done), int(how_many)], \
                  "values": {"property": [values], ..}}``
        """
        def to_dict_list(seq):
            return [
                tag.to_dict()
                for tag in seq
            ]

        def process_property(val):
            if isinstance(val, dict):
                return val

            return to_dict_list(val)

        return {
            "all_set": self.is_all_set(),
            "progress": self.progress(),
            "values": {
                property_name: process_property(
                    getattr(self, property_name) or []
                )
                for property_name in worker_mapping().keys()
            }
        }

    def __eq__(self, obj):
        if not isinstance(obj, RequestInfo):
            return False

        return self.url == obj.url

    def __lt__(self, obj):
        return float(self.creation_ts).__lt__(obj.creation_ts)

    def _p_resolveConflict(self, oldState, savedState, newState):
        """
        This method is used to resolve conflicts in ZODB, when two processes
        create two mismatched transactions. Which actually happens all the
        time.

        Standard ZODB stuff, just google it.
        """
        # for native properties ..
        native_properties = {
            "url",
            "domain",
            # "index",  # this may be big, so don't merge it
            "creation_ts",
            "downloaded_ts",
            "processing_started_ts",
            "processing_ended_ts",
        }

        # .. just pick the biggest of the native properties
        for np_name in native_properties:
            oldState[np_name] = max({savedState[np_name], newState[np_name]})

        # for `mapping` properties perform the merge
        for name in worker_mapping().keys():
            # don't update None states - this is important for .is_all_set()
            if oldState[name] is None and savedState[name] is None and \
               newState[name] is None:
                continue

            # merge
            pool = set()
            if oldState[name]:
                pool.update(oldState[name])
            if savedState[name]:
                pool.update(savedState[name])
            if newState[name]:
                pool.update(newState[name])

            oldState[name] = sorted(list(pool))

        return oldState



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
# Imports =====================================================================
from browser import document


# Functions & classes =========================================================
class ProgressBar(object):
    """
    Generic progressbar controller.
    """
    def __init__(self, tag, whole_tag):
        self.tag = tag
        self.whole_tag = whole_tag
        self.original_message = tag.text

    @staticmethod
    def _compute_percentage(progress_tuple):
        """
        Compute percentage from (done / len(all)) tuple.
        """
        # division by zero..
        if progress_tuple[0] == 0:
            return 0

        return (100 / progress_tuple[1]) * progress_tuple[0]

    def show(self, progress, msg=None):
        """
        Show the progress bar and set it to `progress` tuple or value.

        Args:
            progress (tuple / int / float): Tuple ``(done / len(all))`` or
                the direct percentage value as int / float.
            msg (str, default None): Alternative background description.
        """
        if self.whole_tag.style.display == "none":
            self.whole_tag.style.display = "block"

        # allow either direct percentage value, or (done / len(all)) pairs
        if isinstance(progress, int) or isinstance(progress, float):
            percentage = progress
        else:
            percentage = self.__class__._compute_percentage(progress)

        # toggle animation
        self.tag.class_name = "progress-bar"
        if percentage < 100:
            self.tag.class_name += " progress-bar-striped active"
        else:
            msg = "Hotovo"

        # show percentage in progress bar
        self.tag.aria_valuemin = percentage
        self.tag.style.width = "{}%".format(percentage)

        if msg:
            self.tag.text = msg

    def hide(self):
        """
        Hide the progressbar.
        """
        self.whole_tag.style.display = "none"

    def reset(self):
        """
        Reset the progressbar to 0, hide it and set original text message at
        background.
        """
        self.hide()
        self.tag.class_name = "progress-bar progress-bar-striped active"
        self.tag.aria_valuemin = 0
        self.tag.style.width = "{}%".format(0)
        self.tag.text = self.original_message


URLBar = ProgressBar(
    tag=document["url_progressbar"],
    whole_tag=document["whole_url_progressbar"]
)
ISSNBar = ProgressBar(
    tag=document["issn_progressbar"],
    whole_tag=document["whole_issn_progressbar"]
)
AuthorBar = ProgressBar(
    tag=document["author_progressbar"],
    whole_tag=document["whole_author_progressbar"]
)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
# Imports =====================================================================
from setuptools import setup
from setuptools import find_packages

from docs import getVersion


# Variables ===================================================================
CHANGELOG = open('CHANGELOG.rst').read()
LONG_DESCRIPTION = "\n\n".join([
    open('README.rst').read(),
    open('CONTRIBUTORS.rst').read(),
    CHANGELOG
])


# Actual setup definition =====================================================
setup(
    name='wa-kat',
    version=getVersion(CHANGELOG),
    description="Web page analyzator for czech webarchive.",
    long_description=LONG_DESCRIPTION,
    url='https://github.com/WebArchivCZ/WA-KAT',

    author='Bystroushaak',
    author_email='bystrousak[at]kitakitsune.org',

    classifiers=[
        "Development Status :: 3 - Alpha",

        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",

        "Intended Audience :: Developers",

        "License :: OSI Approved :: MIT License",
    ],
    license='MIT',

    packages=find_packages('src'),
    package_dir={'': 'src'},

    include_package_data=True,
    zip_safe=False,

    scripts=[
        "bin/wa_kat_server.py",
        "bin/wa_kat_build_conspects.py",
        "bin/wa_kat_build_keyword_index.py",
    ],

    install_requires=open("requirements.txt").read().splitlines(),
    extras_require={
        "test": [
            "pytest",
        ],
        "docs": [
            "sphinx",
            "sphinxcontrib-napoleon",
        ]
    }
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
This module provides a collection of classes, which may be used to get/set
values into HTML GUI elements like <input>, <textarea>, <radio> and so on.

This classes may be used as descriptors for properties. They also defines basic
validations.
"""
# Imports =====================================================================
from browser import window
from browser import document

from components import DropdownHandler

from components import InputController

# virtual filesystem / modules provided by REST API
from virtual_fs import periodes


# Functions & classes =========================================================
class StandardElDescriptor(object):
    """
    Descriptor used for standard ``<input>`` element.

    It allows reading / writing of:
        - values
        - dicts (``{"source": .., "val": ..}``), in which case the source is
          highlighted on the right side
        - lists, in which case the element is transformed to typeahead element,
          which works like select
    """
    def __init__(self, el):
        self.el = el

    def __get__(self, obj, objtype):
        return InputController.get_el(self.el)

    def __set__(self, obj, value):
        self._show_error(False)
        InputController.set_el(self.el, value)

    def validate(self):
        if not self.el.required:
            return True

        if InputController.get_el(self.el):
            self._show_error(False)
            return True

        self._show_error(True)
        return False

    def reset(self):
        self.el.value = ""
        self._show_error(False)
        DropdownHandler.hide_dropdown_glyph(self.el.id)

    def _show_error(self, error=True):
        if error:
            self.el.style.border = "2px solid red"
        else:
            self.el.style.border = "0"


class AnnotationDescriptor(StandardElDescriptor):
    """
    Special descriptor used for `Annotation`. This descriptors operates over
    ``<textarea>``, but ignores lines starting with ``--``.

    Accepted datatypes are same as in :class:`StandardElDescriptor`.
    """
    def __get__(self, obj, objtype):
        value = InputController.get_el(self.el)

        active_lines = [
            line.strip()
            for line in value.splitlines()
            if not line.strip().startswith("--")
        ]

        return "\n".join(active_lines)


class PeriodicityDescriptor(StandardElDescriptor):
    """
    Special descriptor used for `periodicity`. If the periodicity is not set,
    the element offers the user list of periodes from ``virtual_fs/``.

    Accepted datatypes are same as in :class:`StandardElDescriptor`.
    """
    def __init__(self):
        super(self.__class__, self).__init__(document["periode"])
        self.__set__(None, periodes.periode_list)

    def __set__(self, obj, value):
        # custom handlers because of custom make_periode_typeahead_tag func
        parent_id = self.el.parent.id
        if "typeahead" not in parent_id.lower():
            parent_id = self.el.parent.parent.id

        if type(value) in [list, tuple] and len(value) > 1:
            window.make_periode_typeahead_tag("#" + parent_id, value)
            DropdownHandler.set_dropdown_glyph(
                self.el.id,
                "glyphicon-menu-down"
            )
            return

        window.destroy_typeahead_tag("#" + parent_id)
        InputController.set_el(self.el, value)


class RadioDescriptor(object):
    """
    Descriptor used to controll radio inputs.
    """
    def __init__(self, name, val_type=bool):
        self.name = name
        self.type = val_type

        self.elements = [
            el
            for el in document.get(selector="input[name='%s']" % name)
        ]
        self.values = {self._get_right_type(el.value) for el in self.elements}

    def _get_right_type(self, val):
        if self.type == bool:
            return val.lower() == "true"

        return self.type(val)

    def __get__(self, obj, objtype):
        for el in self.elements:
            if el.checked:
                return self._get_right_type(el.value)

        return None

    def __set__(self, obj, value):
        def _normalize(val):
            return str(val).lower()

        for el in self.elements:
            if _normalize(el.value) == _normalize(value):
                el.checked = True
                return

        raise ValueError(
            "Can't set the `%s` to `%s`." % (str(el), repr(value))
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: brython (http://brython.info) (like python3)
#
"""
This module contains class for controlling of some of the HTML elements
(<input>, <input class="typeahead">, <textarea>, <select>).
"""
# Imports =====================================================================
from browser import window
from browser import document

from .dropdown_handler import DropdownHandler
from .placeholder_handler import PlaceholderHandler


# Functions & classes =========================================================
class InputController(object):
    """
    Shared behavior for basic input elements.

    This class allows the inputs, textareas and selects to be set using same
    data format of lists of dicts containing values and their sources.
    """
    # this is used to track what kind of elements were added by typeahead
    _set_by_typeahead = set()

    @staticmethod
    def _set_input(el, value):
        """
        Set content of given `el` to `value`.

        Args:
            el (obj): El reference to input you wish to set.
            value (obj/list): Value to which the `el` will be set.
        """
        if isinstance(value, dict):
            el.value = value["val"]
        elif type(value) in [list, tuple]:
            el.value = ", ".join(item["val"] for item in value)
        else:
            el.value = value

    @staticmethod
    def _set_textarea(el, value):
        """
        Set content of given textarea element `el` to `value`.

        Args:
            el (obj): Reference to textarea element you wish to set.
            value (obj/list): Value to which the `el` will be set.
        """
        if isinstance(value, dict):
            el.text = value["val"]
        elif type(value) in [list, tuple]:
            el.text = "\n\n".join(
                "-- %s --\n%s" % (item["source"], item["val"])
                for item in value
            )
        else:
            el.text = value

    @classmethod
    def _set_typeahead(cls, el, value):
        """
        Convert given `el` to typeahead input and set it to `value`.

        This method also sets the dropdown icons and descriptors.

        Args:
            el (obj): Element reference to the input you want to convert to
                typeahead.
            value (list): List of dicts with two keys: ``source`` and ``val``.
        """
        PlaceholderHandler.reset_placeholder_dropdown(el)

        # if there is no elements, show alert icon in glyph
        if not value and not el.value:
            DropdownHandler.set_dropdown_glyph(el.id, "glyphicon-alert")
            return

        # if there is only one element, don't use typeahead, just put the
        # information to the input, set different dropdown glyph and put source
        # to the dropdown
        if len(value) == 1:
            source = value[0]["source"].strip()
            dropdown_el = DropdownHandler.set_dropdown_glyph(
                el.id,
                "glyphicon-eye-open"
            )
            dropdown_content = "<span class='gray_text'>&nbsp;(%s)</span>"

            # save the source to the dropdown menu
            if source:
                dropdown_el.html = dropdown_content % source[::-1]

            el.value = value[0]["val"]
            return

        # get reference to parent element
        parent_id = el.parent.id
        if "typeahead" not in parent_id.lower():
            parent_id = el.parent.parent.id

        if parent_id in cls._set_by_typeahead:
            window.destroy_typeahead_tag("#" + parent_id)

        # if there are multiple elements, put them to the typeahead and show
        # dropdown glyph
        window.make_typeahead_tag("#" + parent_id, value)
        DropdownHandler.set_dropdown_glyph(el.id, "glyphicon-menu-down")
        PlaceholderHandler.set_placeholder_dropdown(el)
        cls._set_by_typeahead.add(parent_id)

    @classmethod
    def _reset_typeaheads(cls):
        """
        Reset all values set by typeahead back to default.
        """
        for el_id in cls._set_by_typeahead:
            window.destroy_typeahead_tag("#" + el_id)

        cls._set_by_typeahead = set()

    @classmethod
    def set_el(cls, el, value):
        """
        Set given `el` tag element to `value`.

        Automatically choose proper method to set the `value` based on the type
        of the `el`.

        Args:
            el (obj): Element reference to the input you want to convert to
                typeahead.
            value (list): List of dicts with two keys: ``source`` and ``val``.
        """
        if not el:
            return

        tag_name = el.elt.tagName.lower()
        if tag_name == "textarea":
            cls._set_textarea(el, value)
        elif tag_name == "input":
            if "typeahead" in el.class_name.lower():
                cls._set_typeahead(el, value)
            else:
                cls._set_input(el, value)
        elif tag_name == "select":
            el.value = value
        else:
            raise ValueError(
                "Setter for %s (%s) not implemented!" % (tag_name, el.id)
            )

    @staticmethod
    def get_el(el):
        """
        Get value of given `el` tag element.

        Automatically choose proper method to set the `value` based on the type
        of the `el`.

        Args:
            el (obj): Element reference to the input you want to convert to
                typeahead.

        Returns:
            str: Value of the object.
        """
        tag_name = el.elt.tagName.lower()
        if tag_name in {"input", "textarea", "select"}:
            return el.value
        else:
            raise ValueError(
                "Getter for %s (%s) not implemented!" % (tag_name, el.id)
            )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
Parse informations about city where the web is located.
"""
#
# Imports =====================================================================
import socket
from urlparse import urlparse

import dhtmlparser
from ipwhois import IPWhois

from shared import parse_meta
from source_string import SourceString


# Functions & classes =========================================================
def get_ip_address(domain):
    """
    Get IP address for given `domain`. Try to do smart parsing.

    Args:
        domain (str): Domain or URL.

    Returns:
        str: IP address.

    Raises:
        ValueError: If can't parse the domain.
    """
    if "://" not in domain:
        domain = "http://" + domain

    hostname = urlparse(domain).netloc

    if not hostname:
        raise ValueError("Can't parse hostname!")

    return socket.gethostbyname(hostname)


def get_html_geo_place_tags(index_page):
    """
    Get `languages` stored in dublin core ``<meta>`` tags.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    return parse_meta(index_page, "geo.placename", "HTML")


def get_whois_tags(ip_address):
    """
    Get list of tags with `address` for given `ip_address`.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    whois = IPWhois(ip_address).lookup()
    nets = whois.get("nets", None)

    if not nets:
        return []

    # parse cities
    cities = [
        net["city"]
        for net in nets
        if net.get("city", None)
    ]

    # parse address tags
    address_list = []
    for net in nets:
        address = net.get("address", None)
        if not address:
            continue

        # filter company name
        if "description" in net and net["description"]:
            address = address.replace(net["description"], "").strip()

        if "\n" in address:
            address = ", ".join(address.splitlines())

        address_list.append(address)

    return [
        SourceString(val, source="Whois")
        for val in set(cities + address_list)
    ]


# def _get_geoip_tag(ip_address):  # TODO: implement geoip
#     pass


def get_place_tags(index_page, domain):  #: TODO geoip to docstring
    """
    Return list of `place` tags parsed from `meta` and `whois`.

    Args:
        index_page (str): HTML content of the page you wisht to analyze.
        domain (str): Domain of the web, without ``http://`` or other parts.

    Returns:
        list: List of :class:`.SourceString` objects.
    """
    ip_address = get_ip_address(domain)
    dom = dhtmlparser.parseString(index_page)

    place_tags = [
        get_html_geo_place_tags(dom),
        get_whois_tags(ip_address),
        # [_get_geo_place_tag(ip_address)],  # TODO: implement geoip
    ]

    return sum(place_tags, [])  # return flattened list



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Interpreter version: python 2.7
#
"""
REST API definitions.
"""
#
# Imports =====================================================================
from os.path import join
from StringIO import StringIO

from bottle import post
from bottle import response
from bottle import HTTPError
from bottle_rest import form_to_params

from ..settings import API_PATH


# REST API modules
import aleph_api
import virtual_fs
import bottle_index
import analyzers_api
from to_output import to_output
from keywords import get_kw_list


# REST API ====================================================================
@post(join(API_PATH, "as_file/<fn:path>"))
@form_to_params(return_json=False)
def download_as_file(fn, data=None):
    """
    Download given `data` as file `fn`. This service exists to allow frontend
    present user with downloadable files.
    """
    if data is None:
        raise HTTPError(500, "This service require POST `data` parameter.")

    response.set_header("Content-Type", "application/octet-stream")
    response.set_header(
        "Content-Disposition",
        'attachment; filename="%s"' % fn
    )

    return StringIO(data)
